// Copyright (C) 2012 Codership Oy <info@codership.com>

// $Id$

#ifndef __gu_mmh3_test__
#define __gu_mmh3_test__

#include <check.h>

extern Suite *gu_mmh3_suite(void);

#endif /* __gu_mmh3_test__ */
