// Copyright (C) 2020 Codership Oy <info@codership.com>

#ifndef __gu_deqmap_test__
#define __gu_deqmap_test__

#include <check.h>

extern Suite *gu_deqmap_suite(void);

#endif /* __gu_deqmap_test__ */
