/* Copyright (C) 2013 Codership Oy <info@codership.com> */
/**
 * @file generic buffer declaration
 *
 * $Id$
 */

#ifndef _GU_BUF_HPP_
#define _GU_BUF_HPP_

#include "gu_buf.h"

namespace gu
{
    typedef struct gu_buf Buf;
}

#endif /* _GU_BUF_HPP_ */
