//
// Copyright (C) 2016 Codership Oy <info@codership.com>
//

#ifndef GU_THREAD_TEST_HPP
#define GU_THREAD_TEST_HPP

#include <check.h>

extern Suite *gu_thread_suite();

#endif // GU_THREAD_TEST_HPP
