/*
 * Copyright (C) 2011 Codership Oy <info@codership.com>
 *
 * $Id$
 */

/*!
 * @file gcs_priv.h Global declarations private to GCS
 */

#ifndef _gcs_priv_h_
#define _gcs_priv_h_

#include "gcs.hpp"

#define GCS_DESYNC_REQ "self-desync"

#endif /* _gcs_priv_h_ */
