/*
 *Copyright (C) 2012-2014 Codership Oy <info@codership.com>
 */

/*! @file Stores some common definitions to be known throughout the modules */

#ifndef COMMON_DEFS_H
#define COMMON_DEFS_H

#define COMMON_BASE_HOST_KEY     "base_host"
#define COMMON_BASE_PORT_KEY     "base_port"
#define COMMON_BASE_PORT_DEFAULT "4567"

#define COMMON_BASE_DIR_KEY      "base_dir"
#define COMMON_BASE_DIR_DEFAULT  "."

#define COMMON_STATE_FILE "grastate.dat"
#define COMMON_VIEW_STAT_FILE "gvwstate.dat"

#endif // COMMON_DEFS_H
