/* Copyright (C) 2013 Codership Oy <info@codership.com>
 *
 * $Id$
 */

#ifndef __gu_vector_test__
#define __gu_vector_test__

#include <check.h>

extern Suite *gu_vector_suite(void);

#endif /* __gu_vector_test__ */
